# other.py | misc. commands
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import itertools
import random
from difflib import get_close_matches

import wikipedia
from discord.ext import commands

import sciolyid.config as config
from sciolyid.core import send_image
from sciolyid.data import (
    aliases,
    all_categories,
    dealias_group,
    groups,
    logger,
    master_id_list,
    meme_list,
    states,
)
from sciolyid.functions import CustomCooldown, build_id_list

# Discord max message length is 2000 characters, leave some room just in case
MAX_MESSAGE = 1950


class Other(commands.Cog):
    def __init__(self, bot):
        self.bot = bot

    @staticmethod
    def broken_join(input_list, max_size: int = MAX_MESSAGE):
        out = []
        temp = ""
        for item in input_list:
            temp += f"{item}\n"
            if len(temp) > max_size:
                out.append(temp)
                temp = ""
        if temp:
            out.append(temp)
        return out

    # Info - Gives image
    @commands.command(
        help=f"- Gives images of specific {config.options['id_type']}", aliases=["i"]
    )
    @commands.check(CustomCooldown(5.0, bucket=commands.BucketType.user))
    async def info(self, ctx, *, arg):
        logger.info("command: info")

        matches = get_close_matches(
            arg.lower(),
            itertools.chain(master_id_list, *aliases.values()),
            n=1,
            cutoff=0.8,
        )
        if matches:
            item = matches[0]

            if item in itertools.chain.from_iterable(aliases.values()):
                logger.info("matched alias! getting item name.")
                item = next(key for key, value in aliases.items() if item in value)

            delete = await ctx.send("Please wait a moment.")
            an = "an" if item.lower()[0] in ("a", "e", "i", "o", "u") else "a"
            await send_image(
                ctx, str(item), message=f"Here's {an} *{item.lower()}* image!"
            )
            await delete.delete()

        else:
            await ctx.send(
                f"{config.options['id_type'][:-1].title()} not found. Are you sure it's on the list?"
            )

    # List command
    @commands.command(help="- DMs the user with the appropriate list.", name="list")
    @commands.check(CustomCooldown(5.0, bucket=commands.BucketType.user))
    async def list_of_items(self, ctx, state_or_group=""):
        logger.info("command: list")

        group_args = set(
            map(
                dealias_group,
                all_categories.intersection(set(state_or_group.lower().split(" "))),
            )
        )
        state_args = set(states.keys()).intersection(
            set(state_or_group.upper().split(" "))
        )
        if not state_args:
            state_args = {config.options["default_state_list"]}

        build = build_id_list(group_args, state_args)
        group_list = sorted(build)
        item_lists = self.broken_join(group_list)

        display_group = list(group_args)
        if not display_group:
            display_group.append(config.options["id_type"])
        elif len(display_group) > 1:
            display_group[-1] = "and " + display_group[-1]

        display_state = list(state_args)
        if len(display_state) > 1:
            display_state[-1] = "and " + display_state[-1]

        if ctx.author.dm_channel is None:
            await ctx.author.create_dm()

        await ctx.author.dm_channel.send(
            f"**{','.join(display_group).capitalize()} in the {','.join(display_state)} list{'s' if len(display_state) > 1 else ''}:**"
        )
        for page in item_lists:
            await ctx.author.dm_channel.send(f"```\n{page}```")

        await ctx.send(
            f"The {','.join(display_state)} list{'s' if len(display_state) > 1 else ''} has **{len(group_list)}** {','.join(display_group)}.\n"
            + "*A full list has been sent to you via DMs.*"
        )

    # Group command - lists groups
    if config.options["id_groups"]:

        @commands.command(
            help="- Prints a list of all available groups.",
            aliases=[
                config.options["category_name"].lower(),
                config.options["category_name"].lower() + "s",
                "group",
                "category",
                "categories",
            ],
        )
        @commands.check(CustomCooldown(8.0, bucket=commands.BucketType.user))
        async def groups(self, ctx):
            logger.info("command: list")

            await ctx.send(f"**Valid Groups**: `{', '.join(map(str, groups.keys()))}`")

    # Wiki command - argument is the wiki page
    @commands.command(help="- Fetch the wikipedia page for any given argument")
    @commands.check(CustomCooldown(5.0, bucket=commands.BucketType.user))
    async def wiki(self, ctx, *, arg):
        logger.info("command: wiki")

        arg = arg.capitalize()

        try:
            page = wikipedia.page(arg, auto_suggest=False)
        except (
            wikipedia.exceptions.DisambiguationError,
            wikipedia.exceptions.PageError,
        ):
            # fall back to suggestion
            try:
                page = wikipedia.page(arg)
            except wikipedia.exceptions.DisambiguationError:
                await ctx.send(
                    "Sorry, that page was not found. Try being more specific."
                )
                return
            except wikipedia.exceptions.PageError:
                await ctx.send("Sorry, that page was not found.")
                return

        await ctx.send(page.url)

    if config.options["meme_file"]:
        # meme command - sends a random item video/gif
        @commands.command(
            help=f"- Sends a funny {config.options['id_type'][:-1]} video/image!"
        )
        @commands.check(
            CustomCooldown(180.0, disable=True, bucket=commands.BucketType.user)
        )
        async def meme(self, ctx):
            logger.info("command: meme")

            if meme_list:
                await ctx.send(random.choice(meme_list))
            else:
                await ctx.send("No memes available :(")

    if config.options["sendas"]:
        # Send command - for testing purposes only
        @commands.command(help="- send command", hidden=True, aliases=["sendas"])
        @commands.is_owner()
        async def send_as_bot(self, ctx, *, args_str):
            logger.info("command: send")

            logger.info(f"args: {args_str}")
            args = args_str.split(" ")

            channel_id = int(args[0])
            try:
                message = args[1:]
            except IndexError:
                await ctx.send("No message provided!")
            channel = self.bot.get_channel(channel_id)
            await channel.send(" ".join(message))
            await ctx.send("Ok, sent!")

    # # Test command - for testing purposes only
    # @commands.command(help="- test command", hidden=True)
    # async def test(self, ctx, *, item):
    #     logger.info("command: test")
    #     await ctx.send(await get_aliases(item))

    # Test command - for testing purposes only
    @commands.command(help="- test command", hidden=True)
    async def error(self, ctx):
        logger.info("command: error")
        await ctx.send(1 / 0)


def setup(bot):
    bot.add_cog(Other(bot))



# __init__.py | package functions
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sciolyid.config as config


def setup(*args, **kwargs):
    if len(args) == 1 and isinstance(args[0], dict):
        kwargs = args[0]

    for option in config.required:
        try:
            config.options[option] = kwargs[option]
        except KeyError as e:
            raise config.BotConfigError(f"Required setup argument {option}") from e

    for option in tuple(config.optional.keys()) + tuple(config.web_optional.keys()):
        try:
            config.options[option] = kwargs[option]
        except KeyError:
            continue

    if config.options["download_func"] is None:
        for option in config.default_image_required:
            try:
                config.options[option] = kwargs[option]
            except KeyError as e:
                raise config.BotConfigError(f"Required argument {option} if no download function provided") from e

    if kwargs.get("web", None):
        if config.options["download_func"] is not None:
            raise config.BotConfigError("Web cannot be set if a custom download function is set")
        for option in config.web_required:
            try:
                config.options[option] = kwargs[option]
            except KeyError as e:
                raise config.BotConfigError(f"Required web setup argument {option}") from e

    if config.options["category_name"]:
        config.options["id_groups"] = True
        config.options["category_name"] = config.options["category_name"].title()
    else:
        config.options["id_groups"] = False

    directory_config_items = (
        "backups_dir",
        "base_image_url",
        "bot_files_dir",
        "data_dir",
        "download_dir",
        "group_dir",
        "log_dir",
        "state_dir",
        "tmp_upload_dir",
        "validation_local_dir",
        "validation_repo_dir",
    )
    for item in directory_config_items:
        if config.options[item] and not config.options[item].endswith("/"):
            config.options[item] += "/"

    bot_files_subdirs = (
        "backups_dir",
        "download_dir",
        "log_dir",
        "tmp_upload_dir",
        "validation_local_dir",
    )
    for item in bot_files_subdirs:
        if config.options[item]:
            config.options[
                item
            ] = f"{config.options['bot_files_dir']}{config.options[item]}"

    data_subdirs = (
        "group_dir",
        "meme_file",
        "prompt_file",
        "state_dir",
        "wikipedia_file",
    )
    for item in data_subdirs:
        if config.options[item]:
            config.options[item] = f"{config.options['data_dir']}{config.options[item]}"

    config.options["id_type"] = config.options["id_type"].lower()
    config.options["short_id_type"] = (
        config.options["short_id_type"] or config.options["id_type"][0]
    )

    if kwargs.get("web", None):
        config.options["hashes_url"] = config.options["hashes_url"] or [
            "https://raw.githubusercontent.com/"
            + "/".join(url.split("/")[-2:]).split(".")[0]
            + f"/master/{path}hashes.csv"
            for url, path in (
                (config.options["github_image_repo_url"], ""),
                (
                    config.options["validation_repo_url"],
                    config.options["validation_repo_dir"],
                ),
            )
        ]  # default hashes_url is https://raw.githubusercontent.com/{user}/{repo}/master/hashes.csv

        config.options["ids_url"] = config.options["ids_url"] or [
            "https://raw.githubusercontent.com/"
            + "/".join(url.split("/")[-2:]).split(".")[0]
            + f"/master/{path}ids.csv"
            for url, path in (
                (config.options["github_image_repo_url"], ""),
                (
                    config.options["validation_repo_url"],
                    config.options["validation_repo_dir"],
                ),
            )
        ]  # default ids_url is https://raw.githubusercontent.com/{user}/{repo}/master/{path}ids.csv

        config.options["commit_url_format"] = config.options["commit_url_format"] or [
            "https://github.com/"
            + "/".join(url.split("/")[-2:]).split(".")[0]
            + "/commit/{id}"
            for url in (
                config.options["github_image_repo_url"],
                config.options["validation_repo_url"],
            )
        ]  # default commit_url_format is https://github.com/{user}/{repo}/commit/{id}

        config.options["verification_server"] = (
            config.options["verification_server"] or config.options["support_server"]
        )


def start():
    import sciolyid.start_bot  # pylint: disable=unused-import,import-outside-toplevel



# skip.py | commands for skipping images
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from discord.ext import commands

from sciolyid.data import database, format_wiki_url, logger
from sciolyid.data_functions import streak_increment
from sciolyid.functions import CustomCooldown


class Skip(commands.Cog):
    def __init__(self, bot):
        self.bot = bot

    # Skip command - no args
    @commands.command(help="- Skip the current image to get a new one", aliases=["sk"])
    @commands.check(CustomCooldown(5.0, bucket=commands.BucketType.channel))
    async def skip(self, ctx):
        logger.info("command: skip")

        current_item = database.hget(f"channel:{ctx.channel.id}", "item").decode(
            "utf-8"
        )
        database.hset(f"channel:{ctx.channel.id}", "item", "")
        database.hset(f"channel:{ctx.channel.id}", "answered", "1")
        if current_item:  # check if there is image
            url = format_wiki_url(ctx, current_item)
            await ctx.send(f"Ok, skipping {current_item.lower()}")
            await ctx.send(url)  # sends wiki page
            streak_increment(ctx, None)  # reset streak
            if database.exists(f"race.data:{ctx.channel.id}"):
                logger.info("auto sending next image")
                group, state, bw = database.hmget(
                    f"race.data:{ctx.channel.id}", ["group", "state", "bw"]
                )
                media = self.bot.get_cog("Media")
                await media.send_pic(ctx, group.decode("utf-8"), state.decode("utf-8"), bw.decode("utf-8"))
        else:
            await ctx.send("You need to ask for an image first!")


def setup(bot):
    bot.add_cog(Skip(bot))



# media.py | commands for getting images
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import random
import string
from typing import Union

from discord.ext import commands

import sciolyid.config as config
from sciolyid.core import send_image
from sciolyid.data import (
    GenericError,
    all_categories,
    database,
    dealias_group,
    logger,
    states,
)
from sciolyid.data_functions import item_setup, session_increment
from sciolyid.functions import (
    CustomCooldown,
    build_id_list,
    check_state_role,
)

IMAGE_MESSAGE = (
    f"*Here you go!* \n**Use `{config.options['prefixes'][0]}pic` again to get a new image of the same {config.options['id_type'][:-1]}, "
    + f"or `{config.options['prefixes'][0]}skip` to get a new {config.options['id_type'][:-1]}. "
    + f"Use `{config.options['prefixes'][0]}check [guess]` to check your answer. "
    + f"Use `{config.options['prefixes'][0]}hint` for a hint.**"
)


class Media(commands.Cog):
    def __init__(self, bot):
        self.bot = bot

    async def _send_race_next_media(self, ctx):
        if database.exists(f"race.data:{ctx.channel.id}"):
            logger.info("auto sending next image")
            group, state, bw = database.hmget(
                f"race.data:{ctx.channel.id}", ["group", "state", "bw"]
            )
            await self.send_pic(
                ctx, group.decode("utf-8"), state.decode("utf-8"), bw.decode("utf-8")
            )

    def error_handle(self, ctx, group_str: str, state_str: str, bw: bool, retries):
        """Return a function to pass to send_pic() as on_error."""

        async def inner(error):
            nonlocal retries

            # skip current item
            database.hset(f"channel:{ctx.channel.id}", "item", "")
            database.hset(f"channel:{ctx.channel.id}", "answered", "1")

            if retries >= 2:  # only retry twice
                await ctx.send("**Too many retries.**\n*Please try again.*")
                await self._send_race_next_media(ctx)
                return

            if isinstance(error, GenericError) and error.code == 100:
                retries += 1
                await ctx.send("**Retrying...**")
                await self.send_pic(ctx, group_str, state_str, bw, retries)
            else:
                await ctx.send("*Please try again.*")
                await self._send_race_next_media(ctx)

        return inner

    @staticmethod
    def increment_item_frequency(ctx, item):
        item_setup(ctx, item)
        database.zincrby("frequency.item:global", 1, string.capwords(item))
        database.zincrby("frequency.item.refresh:global", 1, string.capwords(item))

    async def send_pic(
        self,
        ctx,
        group_str: str,
        state_str: str,
        bw: Union[bool, str] = False,
        retries=0,
    ):
        if isinstance(bw, str):
            bw = bw == "bw"

        logger.info(
            f"{config.options['id_type'][:-1]}: "
            + database.hget(f"channel:{ctx.channel.id}", "item").decode("utf-8")
        )

        currently_in_race = bool(database.exists(f"race.data:{ctx.channel.id}"))
        new_user = database.zscore("users:global", str(ctx.author.id)) < 10

        answered = int(database.hget(f"channel:{ctx.channel.id}", "answered"))
        logger.info(f"answered: {answered}")
        # check to see if previous item was answered
        if answered:  # if yes, give a new item
            session_increment(ctx, "total", 1)

            if not currently_in_race:
                if config.options["id_groups"]:
                    await ctx.send(
                        f"**Recognized arguments:** *Black & White*: `{bw}`, "
                        + f"**{config.options['category_name']}**: `{'None' if group_str == '' else group_str}`, "
                        + f"**Detected State**: `{'None' if state_str == '' else state_str}`"
                    )
                else:
                    await ctx.send(f"**Recognized arguments:** *Black & White*: `{bw}`")

            choices = build_id_list(group_str, state_str)

            if not choices:
                logger.info(f"no {config.options['id_type']} for taxon/state")
                await ctx.send(
                    f"**Sorry, no {config.options['id_type']} could be found for the taxon/state combo."
                    + "**\n*Please try again*"
                )
                return

            if len(choices) < 5:
                logger.info(f"list less than 5 items")
                await ctx.send(
                    f"**Sorry, you must have at least 5 {config.options['id_type']} in the taxon/state combo."
                    + f"**\n*Please try again with more {config.options['id_type']}.*"
                )
                return

            current_item = random.choice(choices)
            self.increment_item_frequency(ctx, current_item)

            prevI = database.hget(f"channel:{ctx.channel.id}", "prevI").decode("utf-8")
            while current_item == prevI and len(choices) > 1:
                current_item = random.choice(choices)
            database.hset(f"channel:{ctx.channel.id}", "prevI", str(current_item))
            database.hset(f"channel:{ctx.channel.id}", "item", str(current_item))
            logger.info("currentItem: " + str(current_item))
            database.hset(f"channel:{ctx.channel.id}", "answered", "0")
            await send_image(
                ctx,
                current_item,
                on_error=self.error_handle(ctx, group_str, state_str, bw, retries),
                message=IMAGE_MESSAGE
                if not currently_in_race and new_user
                else "*Here you go!*",
                bw=bw,
            )
        else:  # if no, give the same item
            await send_image(
                ctx,
                database.hget(f"channel:{ctx.channel.id}", "item").decode("utf-8"),
                on_error=self.error_handle(ctx, group_str, state_str, bw, retries),
                message=IMAGE_MESSAGE
                if not currently_in_race and new_user
                else "*Here you go!*",
                bw=bw,
            )

    @staticmethod
    async def parse(ctx, args_str: str):
        """Parse arguments for options."""

        args = set(args_str.strip().split(" "))
        args.discard("")

        logger.info(f"args: {args}")

        if not database.exists(f"race.data:{ctx.channel.id}"):
            group_args = set()
            state_args = set()
            for arg in args:
                arg = arg.lower()
                if arg == "bw":
                    continue
                if arg in all_categories:
                    group_args.add(dealias_group(arg))
                elif arg.upper() in states.keys():
                    state_args.add(arg.upper())
                else:
                    await ctx.send(f"**Invalid argument provided**: `{arg}`")
                    return None
            group = " ".join(group_args).strip()

            if database.exists(f"session.data:{ctx.author.id}"):
                logger.info("session parameters")

                if group_args:
                    current_groups = set(
                        database.hget(f"session.data:{ctx.author.id}", "group")
                        .decode("utf-8")
                        .split(" ")
                    )
                    logger.info(f"toggle groups: {group_args}")
                    logger.info(f"current groups: {current_groups}")
                    group_args.symmetric_difference_update(current_groups)
                    group_args.discard("")
                    logger.info(f"new groups: {group_args}")
                    group = " ".join(group_args).strip()
                else:
                    group = database.hget(
                        f"session.data:{ctx.author.id}", "group"
                    ).decode("utf-8")

                chosen_state = (
                    database.hget(f"session.data:{ctx.author.id}", "state")
                    .decode("utf-8")
                    .split(" ")
                )
                if chosen_state[0] == "":
                    chosen_state = []
                if not chosen_state:
                    logger.info("no session lists")
                    chosen_state = check_state_role(ctx)

                session_bw = (
                    database.hget(f"session.data:{ctx.author.id}", "bw").decode("utf-8")
                    == "bw"
                )
                bw = not session_bw if "bw" in args else session_bw
            else:
                chosen_state = check_state_role(ctx)
                bw = "bw" in args

            if state_args:
                logger.info(f"toggle states: {state_args}")
                logger.info(f"current states: {chosen_state}")
                state_args.symmetric_difference_update(set(chosen_state))
                state_args.discard("")
                logger.info(f"new states: {state_args}")
                state = " ".join(state_args).strip()
            else:
                state = " ".join(chosen_state).strip()

        else:
            logger.info("race parameters")

            race_bw = (
                database.hget(f"race.data:{ctx.channel.id}", "bw").decode("utf-8")
                == "bw"
            )
            bw = not race_bw if "bw" in args else race_bw

            group = database.hget(f"race.data:{ctx.channel.id}", "group").decode(
                "utf-8"
            )
            state = database.hget(f"race.data:{ctx.channel.id}", "state").decode(
                "utf-8"
            )

        logger.info(f"args: bw: {bw}; group: {group}; state: {state}")

        return (bw, group, state)

    # Pic command - no args
    # help text
    @commands.command(
        help="- Sends a random image for you to ID",
        aliases=["p", config.options["id_type"][:-1], config.options["short_id_type"]],
    )
    # 5 second cooldown
    @commands.check(CustomCooldown(5.0, bucket=commands.BucketType.channel))
    async def pic(self, ctx, *, args_str: str = ""):
        logger.info("command: pic")
        logger.info(f"args: {args_str}")

        data = await self.parse(ctx, args_str)
        if not data:
            return
        bw, group, state = data

        await self.send_pic(ctx, group, state, bw=bw)


def setup(bot):
    bot.add_cog(Media(bot))



# functions.py | function definitions
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import errno
import itertools
import os
import pickle
from typing import Iterable, Union

import aiohttp
import discord
import redis
import wikipedia
from discord.ext import commands
from sentry_sdk import capture_exception

import sciolyid.config as config
import sciolyid.data
from sciolyid.data import (
    GenericError,
    all_categories,
    database,
    dealias_group,
    get_category,
    groups,
    id_list,
    logger,
    states,
)
from sciolyid.data_functions import channel_setup
from sciolyid.util import fetch_get_user


def check_state_role(ctx) -> list:
    """Returns a list of state roles a user has.

    `ctx` - Discord context object
    """
    logger.info("checking roles")
    if not config.options["state_roles"]:
        return []
    user_states = []
    if ctx.guild is not None:
        logger.info("server context")
        user_role_names = [role.name.lower() for role in ctx.author.roles]
        for state in states:
            # gets similarities
            if set(user_role_names).intersection(set(states[state]["aliases"])):
                user_states.append(state)
    else:
        logger.info("dm context")
    logger.info(f"user roles: {user_states}")
    return user_states


async def send_leaderboard(
    ctx, title, page, database_key=None, data=None, items_per_page=10
):
    logger.info("building/sending leaderboard")

    if database_key is None and data is None:
        raise GenericError("database_key and data are both NoneType", 990)
    if database_key is not None and data is not None:
        raise GenericError("database_key and data are both set", 990)

    if page < 1:
        page = 1

    entry_count = (
        int(database.zcard(database_key)) if database_key is not None else data.count()
    )
    page = (page * 10) - 10

    if entry_count == 0:
        logger.info(f"no items in {database_key}")
        await ctx.send("There are no items in the database.")
        return

    if page > entry_count:
        page = entry_count - (entry_count % 10)

    leaderboard_list = (
        map(
            lambda x: (x[0].decode("utf-8"), x[1]),
            database.zrevrangebyscore(
                database_key, "+inf", "-inf", page, items_per_page, True
            ),
        )
        if database_key is not None
        else data.iloc[page : page + items_per_page - 1].items()
    )
    embed = discord.Embed(type="rich", colour=discord.Color.blurple())
    embed.set_author(name=config.options["bot_signature"])
    leaderboard = ""

    for i, stats in enumerate(leaderboard_list):
        leaderboard += f"{i+1+page}. **{stats[0]}** - {int(stats[1])}\n"
    embed.add_field(name=title, value=leaderboard, inline=False)

    await ctx.send(embed=embed)


def build_id_list(
    categories: Union[str, Iterable] = "", state: Union[str, Iterable] = ""
):
    """Generates an ID list based on given arguments

    - `categories`: category string/Iterable
    - `state`: state string/Iterable
    """
    logger.info("building id list")
    if isinstance(categories, str):
        categories = categories.split(" ")
    if isinstance(state, str):
        state = state.split(" ")

    id_choices = set()
    group_args = set(
        map(dealias_group, all_categories.intersection(set(map(str.lower, categories))))
    )
    state_args = set(states.keys()).intersection(set(map(str.upper, state)))
    logger.info(f"group_args: {group_args}, state_args: {state_args}")

    if not config.options["id_groups"]:
        logger.info("no groups allowed")
        group_args = set()

    if group_args:
        items_in_group = set(
            itertools.chain.from_iterable(groups.get(o, []) for o in group_args)
        )
        if state_args:
            items_in_state = set(
                itertools.chain.from_iterable(
                    states[state]["list"] for state in state_args
                )
            )
            id_choices = items_in_group.intersection(items_in_state)
        else:
            id_choices = items_in_group.intersection(set(id_list))
    elif state_args:
        id_choices = set(
            itertools.chain.from_iterable(states[state]["list"] for state in state_args)
        )
    else:
        id_choices = id_list

    logger.info(f"id_choices length: {len(id_choices)}")
    return tuple(id_choices)


def backup_all():
    """Backs up the database to a file.

    This function serializes all data in the REDIS database
    into a file in the `backups` directory.

    This function is run with a task every 6 hours and sends the files
    to a specified discord channel.
    """
    logger.info("Starting Backup")
    logger.info("Creating Dump")
    keys = (key.decode("utf-8") for key in database.keys())
    dump = ((database.dump(key), key) for key in keys)
    logger.info("Finished Dump")
    logger.info("Writing To File")
    try:
        os.mkdir(config.options["backups_dir"])
        logger.info("Created backups directory")
    except FileExistsError:
        logger.info("Backups directory exists")
    with open(config.options["backups_dir"] + "dump.dump", "wb") as f:
        with open(config.options["backups_dir"] + "keys.txt", "w") as k:
            for item, key in dump:
                pickle.dump(item, f)
                k.write(f"{key}\n")
    logger.info("Backup Finished")


async def fools(ctx):
    logger.info(f"holiday check: invoked command: {str(ctx.command)}")
    if str(ctx.command) in ("leaderboard", "missed", "score", "streak", "userscore"):
        embed = discord.Embed(
            type="rich",
            colour=discord.Color.blurple(),
            title=f"{str(ctx.command).title()}",
        )
        embed.set_author(name=config.options["bot_signature"])
        embed.add_field(
            name=f"{str(ctx.command).title()}",
            value="User scores and data have been cleared. We apologize for the inconvenience.",
            inline=False,
        )
        await ctx.send(embed=embed)
        raise GenericError(code=666)
    return True


async def get_all_users(bot):
    logger.info("Starting user cache")
    user_ids = map(int, database.zrangebyscore("users:global", "-inf", "+inf"))
    for user_id in user_ids:
        await fetch_get_user(user_id, bot=bot, member=False)
    logger.info("User cache finished")


async def evict_images():
    """Deletes images for items that have exceeded a certain frequency.

    This prevents images from being stale. If the item frequency has
    been incremented more than `evict_threshold` (e.g. 10) times,
    this function will delete the top `max_evict` (e.g. 3) items.
    """
    logger.info("Updating cached images")

    for item in map(
        lambda x: x.decode(),
        database.zrevrangebyscore(
            "frequency.item.refresh:global",
            "+inf",
            min=config.options["evict_threshold"],
            start=0,
            num=config.options["max_evict"],
        ),
    ):
        database.zadd("frequency.item.refresh:global", {item: 0})
        category = get_category(item)
        await config.options["evict_func"](sciolyid.data, category, item.lower())


class CustomCooldown:
    """Halve cooldown times in DM channels."""

    # Code adapted from discord.py example
    def __init__(
        self,
        per: float,
        disable: bool = False,
        bucket: commands.BucketType = commands.BucketType.channel,
    ):
        """Initialize a custom cooldown.

        `per` (float) - Cooldown default duration, halves in DM channels
        `bucket` (commands.BucketType) - cooldown scope, defaults to channel
        """
        rate = 1
        dm_per = per / 2
        race_per = 0.5
        self.disable = disable
        self.default_mapping = commands.CooldownMapping.from_cooldown(rate, per, bucket)
        self.dm_mapping = commands.CooldownMapping.from_cooldown(rate, dm_per, bucket)
        self.race_mapping = commands.CooldownMapping.from_cooldown(
            rate, race_per, bucket
        )

    def __call__(self, ctx: commands.Context):
        if not self.disable and ctx.guild is None:
            # halve cooldown in DMs
            bucket = self.dm_mapping.get_bucket(ctx.message)

        elif ctx.command.name.startswith("check") and ctx.channel.name.startswith(
            "racing"
        ):
            # tiny check cooldown in racing channels
            bucket = self.race_mapping.get_bucket(ctx.message)

        else:
            bucket = self.default_mapping.get_bucket(ctx.message)

        retry_after = bucket.update_rate_limit()
        if retry_after:
            raise commands.CommandOnCooldown(bucket, retry_after)
        return True


async def handle_error(ctx, error):
    if isinstance(error, commands.CommandOnCooldown):  # send cooldown
        await ctx.send(
            "**Cooldown.** Try again after " + str(round(error.retry_after, 2)) + " s.",
            delete_after=5.0,
        )

    elif isinstance(error, commands.CommandNotFound):
        capture_exception(error)
        await ctx.send("Sorry, the command was not found.")

    elif isinstance(error, commands.MissingRequiredArgument):
        await ctx.send("This command requires an argument!")

    elif isinstance(error, commands.BadArgument):
        await ctx.send("The argument passed was invalid. Please try again.")

    elif isinstance(error, commands.ArgumentParsingError):
        await ctx.send("An invalid character was detected. Please try again.")

    elif isinstance(error, commands.BotMissingPermissions):
        await ctx.send(
            "**The bot does not have enough permissions to fully function.**\n"
            + f"**Permissions Missing:** `{', '.join(map(str, error.missing_perms))}`\n"
            + "*Please try again once the correct permissions are set.*"
        )

    elif isinstance(error, commands.MissingPermissions):
        await ctx.send(
            "You do not have the required permissions to use this command.\n"
            + f"**Required Perms:** `{'`, `'.join(error.missing_perms)}`"
        )

    elif isinstance(error, commands.NoPrivateMessage):
        await ctx.send("**This command is unavailable in DMs!**")

    elif isinstance(error, commands.PrivateMessageOnly):
        await ctx.send("**This command is only available in DMs!**")

    elif isinstance(error, commands.NotOwner):
        logger.info("not owner")
        await ctx.send("Sorry, the command was not found.")

    elif isinstance(error, GenericError):
        if error.code == 192:
            # channel is ignored
            return
        if error.code == 842:
            # user is banned
            return
        if error.code == 666:
            logger.info("GenericError 666")
        elif error.code == 201:
            logger.info("HTTP Error")
            capture_exception(error)
            await ctx.send(
                "**An unexpected HTTP Error has occurred.**\n *Please try again.*"
            )
        else:
            logger.info("uncaught generic error")
            capture_exception(error)
            await ctx.send(
                "**An uncaught generic error has occurred.**\n"
                + "*Please log this message in #support in the support server below, or try again.*\n"
                + f"**Error code:** `{error.code}`"
            )
            await ctx.send(config.options["support_server"])
            raise error

    elif isinstance(error, commands.CommandInvokeError):
        if isinstance(error.original, redis.exceptions.ResponseError):
            capture_exception(error.original)
            if database.exists(f"channel:{ctx.channel.id}"):
                await ctx.send(
                    "**An unexpected ResponseError has occurred.**\n"
                    + "*Please log this message in #support in the support server below, or try again.*\n"
                )
                await ctx.send(config.options["support_server"])
            else:
                await channel_setup(ctx)
                await ctx.send("Please run that command again.")

        elif isinstance(error.original, wikipedia.exceptions.DisambiguationError):
            await ctx.send("Wikipedia page not found. (Disambiguation Error)")

        elif isinstance(error.original, wikipedia.exceptions.PageError):
            await ctx.send("Wikipedia page not found. (Page Error)")

        elif isinstance(error.original, wikipedia.exceptions.WikipediaException):
            capture_exception(error.original)
            await ctx.send("Wikipedia page unavailable. Try again later.")

        elif isinstance(error.original, discord.Forbidden):
            if error.original.code == 50007:
                await ctx.send(
                    "I was unable to DM you. Check if I was blocked and try again."
                )
            elif error.original.code == 50013:
                await ctx.send(
                    "There was an error with permissions. Check the bot has proper permissions and try again."
                )
            else:
                capture_exception(error)
                await ctx.send(
                    "**An unexpected Forbidden error has occurred.**\n"
                    + "*Please log this message in #support in the support server below, or try again.*\n"
                    + f"**Error code:** `{error.original.code}`"
                )
                await ctx.send(config.options["support_server"])

        elif isinstance(error.original, discord.HTTPException):
            capture_exception(error.original)
            if error.original.status == 502:
                await ctx.send(
                    "**An error has occured with discord. :(**\n*Please try again.*"
                )
            else:
                await ctx.send(
                    "**An unexpected HTTPException has occurred.**\n"
                    + "*Please log this message in #support in the support server below, or try again*\n"
                    + f"**Reponse Code:** `{error.original.status}`"
                )
                await ctx.send(config.options["support_server"])

        elif isinstance(error.original, aiohttp.ClientOSError):
            capture_exception(error.original)
            if error.original.errno == errno.ECONNRESET:
                await ctx.send(
                    "**An error has occured with discord. :(**\n*Please try again.*"
                )
            else:
                await ctx.send(
                    "**An unexpected ClientOSError has occurred.**\n"
                    + "*Please log this message in #support in the support server below, or try again.*\n"
                    + "**Error:** "
                    + str(error.original)
                )
                await ctx.send(config.options["support_server"])

        elif isinstance(error.original, aiohttp.ServerDisconnectedError):
            capture_exception(error.original)
            await ctx.send("**The server disconnected.**\n*Please try again.*")

        elif isinstance(error.original, asyncio.TimeoutError):
            capture_exception(error.original)
            await ctx.send("**The request timed out.**\n*Please try again in a bit.*")

        elif isinstance(error.original, OSError):
            capture_exception(error.original)
            if error.original.errno == errno.ENOSPC:
                await ctx.send(
                    "**No space is left on the server!**\n"
                    + "*Please report this message in #support in the support server below!*\n"
                )
                await ctx.send(config.options["support_server"])
            else:
                await ctx.send(
                    "**An unexpected OSError has occurred.**\n"
                    + "*Please log this message in #support in the support server below, or try again.*\n"
                    + f"**Error code:** `{error.original.errno}`"
                )
                await ctx.send(config.options["support_server"])

        else:
            logger.info("uncaught command error")
            capture_exception(error.original)
            await ctx.send(
                "**An uncaught command error has occurred.**\n"
                + "*Please log this message in #support in the support server below, or try again.*\n"
            )
            await ctx.send(config.options["support_server"])
            raise error

    else:
        logger.info("uncaught non-command")
        capture_exception(error)
        await ctx.send(
            "**An uncaught non-command error has occurred.**\n"
            + "*Please log this message in #support in the support server below, or try again.*\n"
        )
        await ctx.send(config.options["support_server"])
        raise error



# __init__.py | pubic module methods
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from sciolyid import setup as _setup


def setup(*args, **kwargs):
    if len(args) == 1 and isinstance(args[0], dict):
        kwargs = args[0]
    kwargs["web"] = True
    _setup(kwargs)


def get_app():
    from sciolyid.web.main import app  # pylint: disable=import-outside-toplevel

    return app



# start_bot.py | main program
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import concurrent.futures
import os
import sys
from datetime import date, datetime, timedelta, timezone

import discord
from discord.ext import commands, tasks

import sciolyid.config as config
import sciolyid.data
from sciolyid.data import GenericError, database, logger
from sciolyid.data_functions import user_setup, channel_setup
from sciolyid.functions import (
    backup_all,
    evict_images,
    fools,
    get_all_users,
    handle_error,
)
from sciolyid.util import prune_user_cache


class CustomBot(commands.Bot):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.on_message_handler = []

    async def on_message(self, message: discord.Message):
        prefixes = await self.get_prefix(message)
        if isinstance(prefixes, str):
            command = message.content.startswith(prefixes)
        else:
            command = message.content.startswith(tuple(prefixes))

        if not message.author.bot and not command:
            for handler in self.on_message_handler:
                await handler(message)
        await super().on_message(message)

    def add_message_handler(self, handler):
        self.on_message_handler.append(handler)


# Initialize bot
intent: discord.Intents = discord.Intents.none()
intent.guilds = True
intent.members = config.options["members_intent"]
intent.messages = True
intent.voice_states = True

cache_flags: discord.MemberCacheFlags = discord.MemberCacheFlags.none()
cache_flags.voice = True
cache_flags.joined = config.options["members_intent"]

bot = CustomBot(
    command_prefix=commands.when_mentioned_or(*config.options["prefixes"]),
    case_insensitive=True,
    description=config.options["bot_description"],
    help_command=commands.DefaultHelpCommand(verify_checks=False),
    intents=intent,
    member_cache_flags=cache_flags,
)


@bot.event
async def on_ready():
    print("Ready!")
    logger.info("Logged in as:")
    logger.info(bot.user.name)
    logger.info(bot.user.id)
    # Change discord activity
    await bot.change_presence(
        activity=discord.Activity(type=3, name=config.options["id_type"])
    )

    # start tasks
    if config.options["refresh_images"]:
        update_images.start()
    if config.options["evict_images"]:
        refresh_images.start()
    refresh_user_cache.start()
    evict_user_cache.start()
    if config.options["backups_channel"]:
        refresh_backup.start()


# Here we load our extensions(cogs) that are located in the cogs directory, each cog is a collection of commands
initial_extensions = [
    "sciolyid.cogs.media",
    "sciolyid.cogs.check",
    "sciolyid.cogs.skip",
    "sciolyid.cogs.hint",
    "sciolyid.cogs.score",
    "sciolyid.cogs.stats",
    "sciolyid.cogs.sessions",
    "sciolyid.cogs.race",
    "sciolyid.cogs.meta",
    "sciolyid.cogs.other",
]
if config.options["state_roles"]:
    initial_extensions.append("sciolyid.cogs.state")

for extension in config.options["disable_extensions"]:
    try:
        initial_extensions.remove(f"sciolyid.cogs.{extension}")
    except ValueError as e:
        raise config.BotConfigError(
            f"Unable to disable extension 'sciolyid.cogs.{extension}'"
        ) from e

initial_extensions += config.options["custom_extensions"]

for extension in set(initial_extensions):
    try:
        bot.load_extension(extension)
    except (
        discord.ClientException,
        ModuleNotFoundError,
        commands.errors.ExtensionFailed,
    ) as e:
        if isinstance(e, commands.errors.ExtensionFailed) and e.args[0].endswith(
            "is already an existing command or alias."
        ):
            raise config.BotConfigError(
                f"short_id_type conflicts with a prexisting command in {extension}"
            )

        raise GenericError(f"Failed to load extension {extension}.", 999) from e


if sys.platform == "win32":
    asyncio.set_event_loop(asyncio.ProactorEventLoop())

######
# Global Command Checks
######


@bot.check
async def prechecks(ctx):
    await ctx.trigger_typing()

    logger.info("global check: checking permissions")
    await commands.bot_has_permissions(
        send_messages=True, embed_links=True, attach_files=True
    ).predicate(ctx)

    logger.info("global check: checking banned")
    if database.zscore("ignore:global", str(ctx.channel.id)) is not None:
        raise GenericError(code=192)
    if database.zscore("banned:global", str(ctx.author.id)) is not None:
        raise GenericError(code=842)

    logger.info("global check: logging command frequency")
    database.zincrby("frequency.command:global", 1, str(ctx.command))

    logger.info("global check: database setup")
    await channel_setup(ctx)
    await user_setup(ctx)

    return True


if config.options["holidays"]:

    @bot.check
    async def is_holiday(ctx):
        """April Fools Prank.

        Can be extended to other holidays as well.
        """
        logger.info("global check: checking holiday")

        if ctx.command.name == "noholiday":
            return True
        if ctx.guild and database.sismember("noholiday:global", str(ctx.guild.id)):
            return True

        now = datetime.now(tz=timezone(-timedelta(hours=4))).date()
        if now == date(now.year, 4, 1):
            return await fools(ctx)
        return True


######
# GLOBAL ERROR CHECKING
######
@bot.event
async def on_command_error(ctx, error):
    """Handles errors for all commands without local error handlers."""
    logger.info("Error: " + str(error))

    # don't handle errors with local handlers
    if hasattr(ctx.command, "on_error"):
        return

    await handle_error(ctx, error)


if config.options["refresh_images"]:

    @tasks.loop(hours=24.0)
    async def update_images():
        """Updates the images."""
        logger.info("updating images")
        await config.options["download_func"](sciolyid.data, None, None)
        logger.info("done updating images!")


if config.options["evict_images"]:

    @tasks.loop(minutes=config.options["evict_frequency"])
    async def refresh_images():
        """Task to delete a random selection of cached images every hour."""
        logger.info("TASK: Refreshing some cache items")
        await evict_images()


@tasks.loop(hours=3.0)
async def refresh_user_cache():
    """Task to update User cache to increase performance of commands."""
    logger.info("TASK: Updating User cache")
    await get_all_users(bot)


@tasks.loop(minutes=8.0)
async def evict_user_cache():
    """Task to remove keys from the User cache to ensure freshness."""
    logger.info("TASK: Removing user keys")
    prune_user_cache(10)


@tasks.loop(hours=1.0)
async def refresh_backup():
    """Sends a copy of the database to a discord channel (BACKUPS_CHANNEL)."""
    logger.info("Refreshing backup")
    try:
        os.remove(config.options["backups_dir"] + "dump.dump")
        logger.info("Cleared backup dump")
    except FileNotFoundError:
        logger.info("Already cleared backup dump")
    try:
        os.remove(config.options["backups_dir"] + "keys.txt")
        logger.info("Cleared backup keys")
    except FileNotFoundError:
        logger.info("Already cleared backup keys")

    event_loop = asyncio.get_event_loop()
    with concurrent.futures.ThreadPoolExecutor(1) as executor:
        await event_loop.run_in_executor(executor, backup_all)

    logger.info("Sending backup files")
    channel = bot.get_channel(config.options["backups_channel"])
    with open(config.options["backups_dir"] + "dump.dump", "rb") as f:
        await channel.send(file=discord.File(f, filename="dump"))
    with open(config.options["backups_dir"] + "keys.txt", "r") as f:
        await channel.send(file=discord.File(f, filename="keys.txt"))
    logger.info("Backup Files Sent!")


# Actually run the bot
token = os.getenv(config.options["bot_token_env"])
bot.run(token)



import os
import pickle
import sys

import redis

if os.getenv("LOCAL_REDIS") == "true":
    database = redis.Redis(host="localhost", port=6379, db=0)
else:
    database = redis.from_url(os.getenv("REDIS_URL"))
folder = sys.argv[1].rstrip("/") if len(sys.argv) > 1 else "backups"


def restore_all():
    print("reading dump")
    with open(f"{folder}/dump.dump", "rb") as f:
        with open(f"{folder}/keys.txt", "r") as k:
            print("restoring")
            for line in k:
                key = line.strip()
                data = pickle.load(f)
                database.restore(key, 0, data, True)
    print("restore finished")


if __name__ == "__main__":
    restore_all()



# data.py | import data from lists
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import csv
import datetime
import itertools
import logging
import logging.handlers
import os
import shutil
import sys

import redis
import sentry_sdk
from discord.ext import commands
from sentry_sdk.integrations.aiohttp import AioHttpIntegration
from sentry_sdk.integrations.redis import RedisIntegration

from sciolyid import config
from sciolyid.downloads import download_github, download_logger

# define database for one connection
if config.options["local_redis"]:
    database = redis.Redis(host="localhost", port=6379, db=0)
elif config.options["redis_env"] is not None:
    database = redis.from_url(os.getenv(config.options["redis_env"]))
else:
    raise ValueError("redis_env must be set if local_redis is False")


def before_sentry_send(event, hint):
    """Fingerprint certain events before sending to Sentry."""
    if "exc_info" in hint:
        error = hint["exc_info"][1]
        if isinstance(error, commands.CommandNotFound):
            event["fingerprint"] = ["command-not-found"]
        elif isinstance(error, commands.CommandOnCooldown):
            event["fingerprint"] = ["command-cooldown"]
    return event


# add sentry logging
if config.options["sentry"]:
    if config.options["sentry_dsn_env"] is None:
        raise ValueError("sentry_dsn_env must be set if sentry is True")
    sentry_sdk.init(
        release=f"Deployed Discord Bot at {datetime.datetime.today()}",
        dsn=os.getenv(config.options["sentry_dsn_env"]),
        integrations=[RedisIntegration(), AioHttpIntegration()],
        before_send=before_sentry_send,
        traces_sample_rate=0.75,
    )

# Database Format Definitions

# prevJ - makes sure it sends a diff image
# prevI - makes sure it sends a diff item (img)

# server format = {
# channel:channel_id : { "item", "answered",
#                     "prevJ", "prevI" }
# }

# session format:
# session.data:user_id : {
#                    "start": 0,
#                    "stop": 0,
#                    "correct": 0,
#                    "incorrect": 0,
#                    "total": 0,
#                    "state": state,
#                    "bw": bw, - Toggles if "bw", doesn't if empty (""), default ""
#                    "group": group,
#                    "wiki": wiki, - Enables if "wiki", disables if empty (""), default "wiki"
#                    "strict": strict - Enables strict spe
# session.incorrect:user_id : [item name, # incorrect]

# race format:
# race.data:ctx.channel.id : {
#                    "start": 0
#                    "stop": 0,
#                    "limit": 10,
#                    "state": state,
#                    "bw": bw,
#                    "group": group,
#                    "strict": strict - Enables strict spelling if "strict", disables if empty, default ""
# }
# race.scores:ctx.channel.id : [ctx.author.id, #correct]

# leaderboard format = {
#    users:global : [user id, # of correct]
#    users.server.id:guild_id : [user id ... ]
# }

# streaks format = {
#    streak:global : [user id, current streak]
#    streak.max:global : [user id, max streak]
# }

# incorrect item format = {
#    incorrect:global : [item name, # incorrect]
#    incorrect.server:guild_id : [item name, # incorrect]
#    incorrect.user:user_id: : [item name, # incorrect]
# }

# correct item format = {
#    correct.user:user_id : [item name, # correct]
# }

# item frequency format = {
#   frequency.item:global : [item name, # displayed]
# }

# last refresh frequency format = {
#   frequency.item.refresh:global : [item name, # displayed]
# }

# command frequency format = {
#   frequency.command:global : [command, # used]
# }

# channel score format = {
#   score:global : [channel id, # of correct]
#   channels:guild_id : [channel id ... ]
# }

# daily update format = {
#     daily.score:YYYY-MM-DD : [user id, # correct today]
#     daily.incorrect:YYYY-MM-DD : [item name, # incorrect today]
# }

# ban format:
#   banned:global : [user id, 0]

# ignore format:
#   ignore:global : [channel id, guild id]

# noholiday format:
#   noholiday:global : { guild id, ... }

# leave confirm format:
#   leave:guild_id : 0

#  states = {
#          state name:
#               {
#               aliases: [alias1, alias2...],
#               list: [item1, item2...],
#               }
#          }

# state items are picked from [state_dir]/[state]/list.txt
# either list can be in any taxon


# setup logging
logger = logging.getLogger(config.options["name"])
if config.options["logs"]:
    os.makedirs(config.options["log_dir"], exist_ok=True)

    file_handler = logging.handlers.TimedRotatingFileHandler(
        f"{config.options['log_dir']}log.txt", backupCount=4, when="midnight"
    )
    file_handler.setLevel(logging.DEBUG)
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.DEBUG)

    file_handler.setFormatter(
        logging.Formatter(
            "{asctime} - {filename:10} -  {levelname:8} - {message}", style="{"
        )
    )
    stream_handler.setFormatter(
        logging.Formatter("{filename:10} -  {levelname:8} - {message}", style="{")
    )

    logger.setLevel(logging.DEBUG)
    logger.addHandler(file_handler)
    logger.addHandler(stream_handler)
    download_logger.setLevel(logging.DEBUG)
    download_logger.addHandler(file_handler)
    download_logger.addHandler(stream_handler)

    # log uncaught exceptions
    def handle_exception(exc_type, exc_value, exc_traceback):
        if issubclass(exc_type, KeyboardInterrupt):
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
            return

        logger.critical(
            "Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback)
        )

    sys.excepthook = handle_exception


class GenericError(commands.CommandError):
    """A custom error class.

    Error codes: (can add more if needed)\n
        0 - no code
        111 - Index Error
        201 - HTTP Error
        999 - Invalid
        990 - Invalid Input
        100 - Blank
        842 - Banned User
        666 - No output error
    """

    def __init__(self, message=None, code=0):
        self.code = code
        super().__init__(message=message)


# Error codes: (can add more if needed)
# 0 - no code
# 111 - Index Error
# 201 - HTTP Error
# 999 - Invalid
# 990 - Invalid Input
# 100 - Blank
# 842 - Banned User
# 666 - No output error


def _wiki_urls():
    logger.info("Working on wiki urls")
    urls = {}
    with open(f'{config.options["wikipedia_file"]}', "r") as f:
        reader = csv.reader(f)
        for item, url in reader:
            urls[item.lower()] = url
    logger.info("Done with wiki urls")
    return urls


# if custom lists are implemented then port https://github.com/tctree333/Bird-ID/pull/290/
def format_wiki_url(ctx, item: str):
    logger.info("fetching wiki url")
    if database.hget(f"session.data:{ctx.author.id}", "wiki") == b"" or database.exists(
        f"race.data:{ctx.channel.id}"
    ):
        logger.info("disabling preview")
        return f"<{wikipedia_urls[item.lower()]}>"
    return wikipedia_urls[item.lower()]


def get_aliases(item: str):
    logger.info(f"getting aliases for {item}")
    item = item.lower()
    try:
        logger.info("aliases found")
        return aliases[item]
    except KeyError:
        logger.info("no aliases")
        return [item]


def get_category(item: str):
    logger.info(f"getting category for item {item}")
    item = item.lower()
    for group in groups:
        if item in groups[group]:
            logger.info(f"category found: {group}")
            return group.lower()
    logger.info(f"no category found for item {item}")
    return None


def dealias_group(group):
    """Resolve group to a real category by expanding aliases."""
    if group not in all_categories:
        return None
    if group in groups.keys():
        return group
    return next(
        key
        for key, value in config.options["category_aliases"].items()
        if group in value
    )


def _groups():
    """Converts txt files of data into lists."""
    filenames = [
        (f"{config.options['group_dir']}{name}", name.lower().split(".")[0])
        for name in os.listdir(config.options["group_dir"])
    ]

    # Converts txt file of data into lists
    lists = {}
    aliases_ = {}
    for filename, category_name in filenames:
        logger.info(f"Working on {filename}")
        lists[category_name] = []
        with open(filename, "r") as f:
            for line in f:
                line = tuple(map(lambda x: x.strip().lower(), line.split(",")))
                lists[category_name].append(line[0])
                if len(line) > 1:
                    aliases_[line[0]] = line
                    logger.info(f"Done with {filename}")

    logger.info("Done with lists!")
    return (lists, aliases_)


def _state_lists():
    """Converts txt files of state data into lists."""
    filenames = ("list", "aliases")
    states_ = {}
    state_names = os.listdir(config.options["state_dir"])
    for state in state_names:
        states_[state.upper()] = {}
        logger.info(f"Working on {state}")
        for filename in filenames:
            logger.info(f"Working on {filename}")
            with open(
                f"{config.options['state_dir']}/{state}/{filename}.txt", "r"
            ) as f:
                states_[state.upper()][filename] = [
                    line.strip().lower() if filename != "aliases" else line.strip()
                    for line in f
                    if line != "EMPTY"
                ]
            logger.info(f"Done with {filename}")
        logger.info(f"Done with {state}")
    logger.info("Done with states list!")
    return states_


def _prompt():
    """Converts a txt file of prompt data into a dictionary of tuples."""
    logger.info("Working on prompts")
    if config.options["prompt_file"]:
        data = {}
        with open(config.options["prompt_file"], "r") as f:
            for line in f.readlines():
                line_content = tuple(map(lambda x: x.strip().lower(), line.split(",")))
                data[line_content[0]] = line_content[1:]
        logger.info("Done with prompts")
        return data
    return {}


def _memes():
    """Converts a txt file of memes/video urls into a list."""
    logger.info("Working on memes")
    if config.options["meme_file"]:
        with open(config.options["meme_file"], "r") as f:
            memes = [line.strip() for line in f]
        logger.info("Done with memes")
        return memes
    return []


def _all_lists():
    """Compiles lists into master lists."""
    logger.info("Working on master lists")
    master = []
    for state in states.values():
        master += state["list"]
    master = tuple(set(master))
    logger.info("Done with master lists!")
    return master


def _config():
    for group in groups:
        if group not in config.options["category_aliases"].keys():
            config.options["category_aliases"][group] = [group]

    _aliases = [
        item for group in groups for item in config.options["category_aliases"][group]
    ]
    if len(_aliases) != len(set(_aliases)):
        raise config.BotConfigError("Aliases in category_aliases not unique")

    if config.options["download_func"] is None:
        config.options["download_func"] = download_github

    if config.options["evict_func"] is None:

        async def evict(_, category, item):
            path = f"{config.options['download_dir']}{category}/{item.lower()}/"
            if os.path.exists(path):
                shutil.rmtree(path)

        config.options["evict_func"] = evict


groups, aliases = _groups()
states = _state_lists()
prompts = _prompt()
meme_list = _memes()
master_id_list = _all_lists()
wikipedia_urls = _wiki_urls()
id_list = states[config.options["default_state_list"]]["list"]
_config()

all_categories = set(
    tuple(groups.keys())
    + tuple(
        item for group in groups for item in config.options["category_aliases"][group]
    )
)  # includes category aliases
possible_words = tuple(
    itertools.chain(master_id_list, *aliases.values(), *prompts.values())
)  # item list that includes item aliases and prompt values

logger.info(f"List Lengths: {len(id_list)}")
logger.info(f"Master List Lengths: {len(master_id_list)}")

logger.info("Done importing data!")



# sessions.py | commands for sessions
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
import time

import discord
from discord.ext import commands

import sciolyid.config as config
from sciolyid.data import all_categories, database, dealias_group, logger, states
from sciolyid.functions import CustomCooldown, check_state_role


class Sessions(commands.Cog):
    def __init__(self, bot):
        self.bot = bot

    @staticmethod
    def _get_options(ctx):
        bw, state, group, wiki, strict = database.hmget(
            f"session.data:{ctx.author.id}", ["bw", "state", "group", "wiki", "strict"]
        )
        options = (
            f"**Black & White:** {bw==b'bw'}\n"
            + (
                f"**{config.options['category_name']}:** {group.decode('utf-8') if group else 'None'}\n"
                if config.options["id_groups"]
                else ""
            )
            + f"**Alternate List:** {state.decode('utf-8') if state else 'None'}\n"
            + f"**Wiki Embeds**: {wiki==b'wiki'}\n"
            + f"**Strict Spelling**: {strict==b'strict'}"
        )

        return options

    @staticmethod
    def _get_stats(ctx):
        start, correct, incorrect, total = map(
            int,
            database.hmget(
                f"session.data:{ctx.author.id}",
                ["start", "correct", "incorrect", "total"],
            ),
        )
        elapsed = str(datetime.timedelta(seconds=round(time.time()) - start))
        try:
            accuracy = round(100 * (correct / (correct + incorrect)), 2)
        except ZeroDivisionError:
            accuracy = 0

        stats = (
            f"**Duration:** `{elapsed}`\n"
            + f"**# Correct:** {correct}\n"
            + f"**# Incorrect:** {incorrect}\n"
            + f"**Total:** {total}\n"
            + f"**Accuracy:** {accuracy}%\n"
        )
        return stats

    async def _send_stats(self, ctx, preamble):
        database_key = f"session.incorrect:{ctx.author.id}"

        embed = discord.Embed(
            type="rich", colour=discord.Color.blurple(), title=preamble
        )
        embed.set_author(name=config.options["bot_signature"])

        if database.zcard(database_key) != 0:
            leaderboard_list = database.zrevrangebyscore(
                database_key, "+inf", "-inf", 0, 5, True
            )
            leaderboard = ""

            for i, stats in enumerate(leaderboard_list):
                leaderboard += (
                    f"{i+1}. **{stats[0].decode('utf-8')}** - {int(stats[1])}\n"
                )
        else:
            logger.info(f"no items in {database_key}")
            leaderboard = f"**There are no missed {config.options['id_type']}.**"

        embed.add_field(name="Options", value=self._get_options(ctx), inline=False)
        embed.add_field(name="Stats", value=self._get_stats(ctx), inline=False)
        embed.add_field(
            name=f"Top Missed {config.options['id_type'].title()}",
            value=leaderboard,
            inline=False,
        )

        await ctx.send(embed=embed)

    @commands.group(
        brief=f"- Base session command. Use '{config.options['prefixes'][0]}help session' for more info.",
        help="- Base session command\nSessions will record your activity for an amount of time and "
        + "will give you stats on how your performance and "
        + "also set global variables such as black and white"
        + (", specific categories, " if config.options["id_groups"] else "")
        + "or alternate lists.",
        aliases=["ses", "sesh"],
    )
    async def session(self, ctx):
        if ctx.invoked_subcommand is None:
            await ctx.send(
                "**Invalid subcommand passed.**\n*Valid Subcommands:* `start, view, stop`"
            )

    # starts session
    @session.command(
        brief="- Starts session",
        help="- Starts session.\n"
        + f"Arguments passed will become the default arguments to '{config.options['prefixes'][0]}{config.options['id_type'][:-1]}', "
        + "but can be manually overwritten during use.\n"
        + f"These settings can be changed at any time with '{config.options['prefixes'][0]}session edit', "
        + "and arguments can be passed in any order.\n",
        aliases=["st"],
        usage=(f"[bw] [state]{' [category]' if config.options['id_groups'] else ''}"),
    )
    @commands.check(CustomCooldown(3.0, bucket=commands.BucketType.user))
    async def start(self, ctx, *args):
        logger.info("command: start session")

        if database.exists(f"session.data:{ctx.author.id}"):
            logger.info("already session")
            await ctx.send(
                f"**There is already a session running.** *Change settings/view stats with `{config.options['prefixes'][0]}session edit`*"
            )
            return

        logger.info(f"args: {args}")

        # parse args
        group_args = set()
        state_args = set()
        bw = ""
        strict = ""
        wiki = ""
        for arg in set(args):
            arg = arg.lower()
            if arg == "bw":
                bw = "bw"
            elif arg == "wiki":
                wiki = "wiki"
            elif arg == "strict":
                strict = "strict"
            elif arg in all_categories:
                group_args.add(dealias_group(arg))
            elif arg.upper() in states.keys():
                state_args.add(arg.upper())
            else:
                await ctx.send(f"**Invalid argument provided**: `{arg}`")
                return

        if state_args:
            state = " ".join(state_args).strip()
        else:
            state = " ".join(check_state_role(ctx))

        if group_args and config.options["id_groups"]:
            group = " ".join(group_args).strip()
        else:
            group = ""

        logger.info(f"adding bw: {bw}; group: {group}; state: {state}; wiki: {wiki}; strict: {strict}")

        database.hset(
            f"session.data:{ctx.author.id}",
            mapping={
                "start": round(time.time()),
                "stop": 0,
                "correct": 0,
                "incorrect": 0,
                "total": 0,
                "bw": bw,
                "state": state,
                "group": group,
                "wiki": wiki,
                "strict": strict,
            },
        )
        await ctx.send(f"**Session started with options:**\n{self._get_options(ctx)}")

    # views session
    @session.command(
        brief="- Views session",
        help="- Views session\nSessions will record your activity for an amount of time and "
        + "will give you stats on how your performance and "
        + "also set global variables such as black and white"
        + (", specific categories, " if config.options["id_groups"] else "")
        + "or alternate lists.",
        aliases=["view"],
        usage=(f"[bw] [state]{' [category]' if config.options['id_groups'] else ''}"),
    )
    @commands.check(CustomCooldown(3.0, bucket=commands.BucketType.user))
    async def edit(self, ctx, *args):
        logger.info("command: view session")

        if not database.exists(f"session.data:{ctx.author.id}"):
            await ctx.send(
                f"**There is no session running.** *You can start one with `{config.options['prefixes'][0]}session start`*"
            )
            return

        logger.info(f"args: {args}")

        # parse args
        group_args = set()
        state_args = set()
        for arg in set(args):
            arg = arg.lower()
            if arg == "bw":
                if not database.hget(f"session.data:{ctx.author.id}", "bw"):
                    logger.info("adding bw")
                    database.hset(f"session.data:{ctx.author.id}", "bw", "bw")
                else:
                    logger.info("removing bw")
                    database.hset(f"session.data:{ctx.author.id}", "bw", "")
            elif arg == "wiki":
                if database.hget(f"session.data:{ctx.author.id}", "wiki"):
                    logger.info("disabling wiki embeds")
                    database.hset(f"session.data:{ctx.author.id}", "wiki", "")
                else:
                    logger.info("enabling wiki embeds")
                    database.hset(f"session.data:{ctx.author.id}", "wiki", "wiki")
            elif arg == "strict":
                if database.hget(f"session.data:{ctx.author.id}", "strict"):
                    logger.info("disabling strict spelling")
                    database.hset(f"session.data:{ctx.author.id}", "strict", "")
                else:
                    logger.info("enabling strict spelling")
                    database.hset(
                        f"session.data:{ctx.author.id}", "strict", "strict"
                    )
            elif arg in all_categories:
                group_args.add(dealias_group(arg))
            elif arg.upper() in states.keys():
                state_args.add(arg.upper())
            else:
                await ctx.send(f"**Invalid argument provided**: `{arg}`")
                return

        if state_args:
            current_states = set(
                database.hget(f"session.data:{ctx.author.id}", "state")
                .decode("utf-8")
                .split(" ")
            )
            logger.info(f"toggle states: {state_args}")
            logger.info(f"current states: {current_states}")
            state_args.symmetric_difference_update(current_states)
            state_args.discard("")
            logger.info(f"new states: {state_args}")
            database.hset(
                f"session.data:{ctx.author.id}",
                "state",
                " ".join(state_args).strip(),
            )

        if group_args and config.options["id_groups"]:
            current_group = set(
                database.hget(f"session.data:{ctx.author.id}", "group")
                .decode("utf-8")
                .split(" ")
            )
            logger.info(f"toggle group: {group_args}")
            logger.info(f"current group: {current_group}")
            group_args.symmetric_difference_update(current_group)
            group_args.discard("")
            logger.info(f"new groups: {group_args}")
            database.hset(
                f"session.data:{ctx.author.id}",
                "group",
                " ".join(group_args).strip(),
            )
        await self._send_stats(ctx, "**Session started previously.**\n")


    # stops session
    @session.command(help="- Stops session", aliases=["stp", "end"])
    @commands.check(CustomCooldown(3.0, bucket=commands.BucketType.user))
    async def stop(self, ctx):
        logger.info("command: stop session")

        if database.exists(f"session.data:{ctx.author.id}"):
            database.hset(f"session.data:{ctx.author.id}", "stop", round(time.time()))

            await self._send_stats(ctx, "**Session stopped.**\n")
            database.delete(f"session.data:{ctx.author.id}")
            database.delete(f"session.incorrect:{ctx.author.id}")
        else:
            await ctx.send(
                f"**There is no session running.** *You can start one with `{config.options['prefixes'][0]}session start`*"
            )


def setup(bot):
    bot.add_cog(Sessions(bot))



# state.py | commands for alternate lists
# Copyright (C) 2019-2021  EraserBird, person_v1.32

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import string

import discord
from discord.ext import commands

import sciolyid.config as config
from sciolyid.data import logger, states
from sciolyid.functions import CustomCooldown, handle_error


class States(commands.Cog):
    def __init__(self, bot):
        self.bot = bot

    async def broken_send(self, ctx, message, between=""):
        pages = []
        temp_out = ""
        for line in message.splitlines(keepends=True):
            temp_out += line
            if len(temp_out) > 1700:
                temp_out = f"{between}{temp_out}{between}"
                pages.append(temp_out.strip())
                temp_out = ""
        temp_out = f"{between}{temp_out}{between}"
        if temp_out.strip():
            pages.append(temp_out.strip())
        for item in pages:
            await ctx.send(item)

    # set state role
    @commands.command(
        help=f"- Sets an alternate {config.options['id_type']} list",
        name="set",
        aliases=["state", "alt"],
    )
    @commands.check(CustomCooldown(5.0, bucket=commands.BucketType.user))
    @commands.guild_only()
    @commands.bot_has_permissions(manage_roles=True)
    async def state(self, ctx, *, args):
        logger.info("command: state set")

        raw_roles = ctx.author.roles
        role_ids = [role.id for role in raw_roles]
        role_names = [role.name.lower() for role in ctx.author.roles]
        args = args.upper().split(" ")

        added = []
        removed = []
        invalid = []
        for arg in args:
            if arg not in states:
                logger.info("invalid state")
                invalid.append(arg)

            # gets similarities
            elif not set(role_names).intersection(set(states[arg]["aliases"])):
                # need to add role (does not have role)
                logger.info("add roles")
                raw_roles = ctx.guild.roles
                guild_role_names = [role.name.lower() for role in raw_roles]
                guild_role_ids = [role.id for role in raw_roles]

                if states[arg]["aliases"][0].lower() in guild_role_names:
                    # guild has role
                    index = guild_role_names.index(states[arg]["aliases"][0].lower())
                    role = ctx.guild.get_role(guild_role_ids[index])

                else:
                    # create role
                    logger.info("creating role")
                    role = await ctx.guild.create_role(
                        name=string.capwords(states[arg]["aliases"][0]),
                        permissions=discord.Permissions.none(),
                        hoist=False,
                        mentionable=False,
                        reason=f"Create role for alternate {config.options['id_type']} list",
                    )

                await ctx.author.add_roles(
                    role,
                    reason=f"Set role for alternate {config.options['id_type']} list",
                )
                added.append(role.name)

            else:
                # have roles already (there were similarities)
                logger.info("already has role, removing")
                index = role_names.index(states[arg]["aliases"][0].lower())
                role = ctx.guild.get_role(role_ids[index])
                await ctx.author.remove_roles(
                    role,
                    reason=f"Remove role for alternate {config.options['id_type']} list",
                )
                removed.append(role.name)

        await ctx.send(
            (
                f"**Sorry,** `{'`, `'.join(invalid)}` **{'are' if len(invalid) > 1 else 'is'} not a valid list.**\n"
                + f"*Valid Lists:* `{'`, `'.join(states.keys())}`\n"
                if invalid
                else ""
            )
            + (
                f"**Added the** `{'`, `'.join(added)}` **role{'s' if len(added) > 1 else ''}**\n"
                if added
                else ""
            )
            + (
                f"**Removed the** `{'`, `'.join(removed)}` **role{'s' if len(removed) > 1 else ''}**\n"
                if removed
                else ""
            )
        )

    @state.error
    async def set_error(self, ctx, error):
        logger.info("state set error")
        if isinstance(error, commands.MissingRequiredArgument):
            await ctx.send(
                f"**Please enter an alternate list.**\n*Valid Lists:* `{'`, `'.join(states.keys())}`"
            )
        else:
            await handle_error(ctx, error)


def setup(bot):
    bot.add_cog(States(bot))



# hint.py | commands for giving hints
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from discord.ext import commands

from sciolyid.data import database, logger
from sciolyid.functions import CustomCooldown


class Hint(commands.Cog):
    def __init__(self, bot):
        self.bot = bot

    # give hint
    @commands.command(
        help="- Gives first letter of current image",
        usage="[count|c last|l all|a]",
        aliases=["h"],
    )
    @commands.check(CustomCooldown(3.0, bucket=commands.BucketType.channel))
    async def hint(self, ctx, *args):
        logger.info("command: hint")

        current_item = database.hget(f"channel:{ctx.channel.id}", "item").decode(
            "utf-8"
        )
        if current_item != "":  # check if there is item
            if len(args) == 0:
                await ctx.send(f"The first letter is {current_item[0]}.")
            elif args[0] == "count" or args[0] == "c":
                await ctx.send(f"The answer has {str(len(current_item))} letters.")
            elif args[0] == "last" or args[0] == "l":
                await ctx.send(f"The last letter is {current_item[-1]}.")
            elif args[0] == "all" or args[0] == "a":
                the_hint = "`" + current_item[0]
                for letter in current_item[1:-1]:
                    if letter != " ":
                        the_hint += " _ "
                    else:
                        the_hint += "   "
                await ctx.send(the_hint + current_item[-1] + "`")
            else:
                await ctx.send(f"The first letter is {current_item[0]}.")
        else:
            await ctx.send("You need to ask for a image first!")


def setup(bot):
    bot.add_cog(Hint(bot))
